# Copyright (C) 2008-2011 Dejan Muhamedagic <dmuhamedagic@suse.de>
# Copyright (C) 2013 Kristoffer Gronlund <kgronlund@suse.com>
# See COPYING for license information.
'''
Session-only options (not saved).
'''

interactive = False
batch = False
ask_no = False
regression_tests = False
profile = ""
history = "live"
input_file = ""
shadow = ""
scriptdir = ""
# set to true when completing non-interactively
shell_completion = False
